/*
* Copyright (C) 2013 - 2016  Xilinx, Inc.  All rights reserved.
*
* Permission is hereby granted, free of charge, to any person
* obtaining a copy of this software and associated documentation
* files (the "Software"), to deal in the Software without restriction,
* including without limitation the rights to use, copy, modify, merge,
* publish, distribute, sublicense, and/or sell copies of the Software,
* and to permit persons to whom the Software is furnished to do so,
* subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included
* in all copies or substantial portions of the Software.
*
* Use of the Software is limited solely to applications:
* (a) running on a Xilinx device, or (b) that interact
* with a Xilinx device through a bus or interconnect.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL XILINX  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*
* Except as contained in this notice, the name of the Xilinx shall not be used
* in advertising or otherwise to promote the sale, use or other dealings in this
* Software without prior written authorization from Xilinx.
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <xparameters.h>
#include <xil_types.h>

#include "dt_fb.h"
#include "hdmio.h"
#include "vdma.h"


#define MAP_SIZE 4096UL
#define MAP_MASK (MAP_SIZE - 1)

typedef struct {
    void *base;
    void *dev_base;
} mem_map_t;

int mapio(int memfd, off_t dev_base, mem_map_t *m)
{
    // Map one page of memory into user space such that the device is in that page, but it may not
    // be at the start of the page.

    m->base = mmap(0, MAP_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, dev_base & ~MAP_MASK);
    if ( m->base == MAP_FAILED ) {  // MAP_FAILED = (void *) -1
        return XST_FAILURE;
    }
    // printf("Memory mapped at address %p.\n", m->base);

    // get the address of the device in user space which will be an offset from the base
    // that was mapped as memory is mapped at the start of a page

    m->dev_base = (void*)((char*)(m->base) + (dev_base & MAP_MASK));

    return XST_SUCCESS;
}


int main(int argc, char **argv)
{
    int status = XST_SUCCESS;

    dt_fb_params_t fb_params;

    (void)get_dt_fb_params("/sys/firmware/devicetree/base/fbxserver0", &fb_params);
    
//     if ( argc < 2 ) {
//         printf("Usage: %s FB_BASE_ADDR_HEX\n", argv[0]);
//         printf("    FB_BASE_ADDR_HEX - base address of the frame buffer (0x2800000 for instance)\n");
//         return -1;
//     }
//     
//     unsigned int fb_base;    
//     sscanf(argv[1], "%x", &fb_base);

    printf("TE0701 - Initialize VDMA for HDMI Out at 0x%X ... \n", fb_params.base_low);      

    
    /*
     * open memory device
     */
    int memfd;
    memfd = open("/dev/mem", O_RDWR | O_SYNC);
    if (memfd < 0) {
        printf("  Can't open /dev/mem. (%d)\n", memfd);
        goto END_ERR;
    }

    
    /*
     * Initialize VDMA
     */
    printf("  VDMA Common init ... ");

    
    XAxiVdma_Config *Config;

    // get basic configuration from the xparameters
    Config = XAxiVdma_LookupConfig( XPAR_VDMA_AXI_VDMA_0_DEVICE_ID );
    if ( !Config ) {
        printf("XAxiVdma_LookupConfig failure\r\n");
        goto END_ERR;
    }

    mem_map_t vdma_map;
    if ( mapio(memfd, Config->BaseAddress, &vdma_map) ) {
        printf("VDMA mapping to userspace failed\n");
        goto END_ERR;
    }
    
    /*
     * update baseAddress pointer after mapping
     */
    Config->BaseAddress = (UINTPTR) vdma_map.dev_base;
    


//     status = vdma_common_init( XPAR_VDMA_AXI_VDMA_0_DEVICE_ID );
    status = vdma_common_init( Config );
    if ( status != XST_SUCCESS ) {
        printf("FAILED\n");
        goto END_ERR;
    } else {
        printf("OK\n");
    }

    /*
     * Configure VDMA Output to 1080p
     */
    printf("  Set VDMA out to %dx%d@60 ... ", fb_params.width, fb_params.height);
//     status = vdma_out_init(fb_base, 1920, 1080, 4);
    status = vdma_out_init(fb_params.base_low, 
                           fb_params.width, 
                           fb_params.height, 
                           fb_params.stride/fb_params.width);
    if ( status != XST_SUCCESS ) {
        printf("FAILED\n");
        goto END_ERR;
    } else {
        printf("OK\n");
    }
    

    if ( munmap(vdma_map.base, MAP_SIZE) == -1 ) {
        printf("I2C controller for HDMI OUT userspace unmapping failed\n");
        goto END_ERR;
    }
    

    if ( close(memfd) ) {
        printf("Cannot close /dev/mem\n");
        goto END_ERR;
    }
        
    printf("Done\n");
    return 0;

END_ERR:
    printf("FAILED\n");
    printf("                                                                              \n\n");
    return status;
}
